﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Reactive.Linq;
using Microsoft.Extensions.Configuration;
using DeviceModelLib;
using DashboardApp2.Hubs;

namespace DashboardApp2
{
    public class InterHubHelper
    {
        class GroupInfoEx : GroupInfo
        {
            public string RuleForDevice { get; set; }
            
            public bool IsOfThisGroup(int deviceId)
            {
                return idSpan[0] <= deviceId && deviceId <= idSpan[1];
            }

            private int[] idSpan = new int[2];

            public void DetermineIdSpan()
            {
                if (string.IsNullOrEmpty(RuleForDevice))
                    return;

                var ss = RuleForDevice.Replace(" ", string.Empty).Split(',');
                if (ss.Length == 2)
                {
                    idSpan[0] = int.Parse(ss[0]);
                    idSpan[1] = int.Parse(ss[1]);
                }
            }
        }

        static InterHubHelper helper;

        public int StreamingIntervalInMs { get; private set; }

        List<DeviceTypeInfo> lstDeviceTypeInfo = new List<DeviceTypeInfo>();
        List<DeviceInfo> lstDeviceInfo = new List<DeviceInfo>();
        List<Measurement> lstMeasurement = new List<Measurement>();
        int historyDepthInSec;
        GroupInfoEx[] groups;
        List<WebHub> lstWebHub = new List<WebHub>();
        Timer timer;
        DateTime lastTime = DateTime.MinValue;
           
        object locker = new object();

        Measurement[] measurements;
        public Measurement[] MeasurementsObj
        {
            get
            {
                lock (locker)
                    return measurements;
            }
        }

        public static InterHubHelper Instance
        {
            get
            {
                if (helper == null)
                    helper = new InterHubHelper();

                return helper;
            }
        }

        private InterHubHelper()
        {
        }

        #region Add / Remove Hub

        public void AddWebHub(WebHub webHub)
        {
            lock (locker)
                lstWebHub.Add(webHub);
        }

        public void RemoveWebHub(WebHub webHub)
        {
            lock (locker)
                lstWebHub.Remove(webHub);
        }

        #endregion Add / Remove Hub

        public void Configure(IConfiguration configuration)
        {
            var configSectionProcessing = configuration.GetSection("Processing");
            StreamingIntervalInMs = configSectionProcessing.GetValue<int>("StreamingIntervalInMs");
            historyDepthInSec = configSectionProcessing.GetValue<int>("HistoryDepthInSec");
            
            var lstGroupEx = new List<GroupInfoEx>();
            int index = -1;
            while (true)
            {
                var item = new GroupInfoEx();
                index++;
                var section = configuration.GetSection($"GroupsInfo:Groups:{index}");
                if (section != null)
                {
                    var arr = section.AsEnumerable().ToArray();
                    var props = item.GetType().GetProperties();
                    var names = props.Select(p => p.Name).ToArray();
                    var values = new List<string>();

                    foreach (var name in names)
                    {
                        var v = arr.Where(a => a.Key.Contains(name)).FirstOrDefault();
                        if (!string.IsNullOrEmpty(v.Value))
                        {
                            var property = props.Where(p => p.Name == name).First();
                            object ob = v.Value;
                            if (property.PropertyType == typeof(int))
                                ob = int.Parse(v.Value);
                            property.SetValue(item, ob);
                        }
                    }

                    if (item.Id > 0)
                    {
                        item.DetermineIdSpan();
                        lstGroupEx.Add(item);
                    }
                    else
                        break;
                }
            }

            groups = lstGroupEx.ToArray();

            // Timer handler
            timer = new Timer(_ =>
               {
                   lock (locker)
                   {
                       if (lastTime > DateTime.MinValue)
                       {
                            // Output only measurements received since last update for this connection
                            measurements = lstMeasurement.Where(m => m.Timestamp > lastTime).ToArray();

                            // Forget old measurements
                            lstMeasurement = lstMeasurement.Where(m => m.Timestamp > DateTime.Now - TimeSpan.FromSeconds(historyDepthInSec)).ToList();
                       }

                       lastTime = DateTime.Now;
                   }
               },
               null, StreamingIntervalInMs, StreamingIntervalInMs);
        }

        public void AddDeviceType(DeviceTypeInfo deviceTypeInfo)
        {
            lock (locker)
            {
                // Check whether device's type has been already registered
                if (lstDeviceTypeInfo.Where(t => t.Id == deviceTypeInfo.Id).Count() == 0)
                    lstDeviceTypeInfo.Add(deviceTypeInfo);
            }
        }

        public void AddDevice(DeviceInfo deviceInfo)
        {
            AssignGroupToDevice(deviceInfo);
            lock (locker)
            {
                // Check whether device has been already registered
                if (lstDeviceInfo.Where(d => d.Id == deviceInfo.Id).Count() == 0)
                {
                    lstDeviceInfo.Add(deviceInfo);
                    foreach (var webHub in lstWebHub)
                    {
                        try
                        {
                            webHub.RegisterDeviceWithWeb(deviceInfo);
                        }
                        catch { }
                    }
                }
            }
         }

        public DeviceTypeInfo[] GetDeviceTypes()
        {
            lock (locker)
                return lstDeviceTypeInfo.ToArray();
        }

        public DeviceInfo[] GetDevices()
        {
            lock (locker)
                return lstDeviceInfo.ToArray();
        }

        public GroupInfo[] GetGroups()
        {
            return groups.Select(g => (GroupInfo)g).ToArray();
        }

        private void AssignGroupToDevice(DeviceInfo deviceInfo)
        {
            var group = groups.Where(g => g.IsOfThisGroup(deviceInfo.Id)).FirstOrDefault();
            if (group != null)
                deviceInfo.GroupId = group.Id;
        }

        public void AddMeasurement(Measurement measurement)
        {
            lock (locker)
                lstMeasurement.Add(measurement);
        }
    }
}